// exemple de composition à plusieurs composant
// cette exemple montre aussi que Rust
// empêche des incohérences : à la compilation
// tout tentatif d'accéder à une référence qui
// ne référence plus rien est détecté.
#[derive(Debug, Clone)]
pub struct Room {
    pub kind: String,
    pub square_meters: usize,
}

#[derive(Default)]
pub struct House {
    // composition à plusieurs composants
    rooms: Vec<Room>,
}

impl House {
    pub fn new() -> Self {
        Self { rooms: Vec::new() }
    }

    pub fn add(
        &mut self,
        kind: String,
        square_meters: usize,
    ) {
        self.rooms.push(Room {
            kind,
            square_meters,
        });
    }

    pub fn get_bedroom(&self) -> Option<&Room> {
        self.rooms.first()
    }
}

// ceci permet un affichage lorsque un objet de type
// House est explicitement détruit (en appelant drop())
impl Drop for House {
    fn drop(&mut self) {
        println!("detroying house");
    }
}

// classes utilitaires pour les exemples suivants
#[derive(Debug, Clone)]
pub struct Color {
    pub red: u8,
    pub green: u8,
    pub blue: u8,
}

#[derive(Debug, Clone)]
pub struct Point {
    pub x: f64,
    pub y: f64,
}

impl Point {
    pub fn translate(
        &mut self,
        dx: f64,
        dy: f64,
    ) {
        self.x += dx;
        self.y += dy;
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// exemple de composition à 1 composant et à 0..1 composant
pub struct Rectangle {
    // composition : 1 composant
    position: Point,
    color: Color,
    // composition : 0..1 composant
    frame: Option<Color>,
    width: f64,
    height: f64,
}

impl Rectangle {
    pub fn new(
        position: Point,
        color: Color,
        frame: Option<Color>,
        width: f64,
        height: f64,
    ) -> Self {
        Self {
            position,
            color,
            frame,
            width,
            height,
        }
    }

    pub fn translate(
        &mut self,
        dx: f64,
        dy: f64,
    ) {
        self.position.translate(dx, dy);
    }

    pub fn draw(&self) {
        println!("\tposition: ({},{})", self.position.x, self.position.y);
        println!("\tdimensions: {}x{}", self.width, self.height);
        println!(
            "\tcolor: [{},{},{}]",
            self.color.red, self.color.green, self.color.blue
        );
        if let Some(f) = &self.frame {
            println!("\tcolor: [{},{},{}]", f.red, f.green, f.blue);
        } else {
            println!("\tNo frame");
        }

        //OU
        /*
        match &self.frame {
            Some(f) => {
                println!("\tcolor: [{},{},{}]", f.red, f.green, f.blue);
            }
            None => {
                println!("\tNo frame");
            }
        }*/
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// exemple composition à plusieurs composants
pub struct Palette {
    size: usize,
    next: usize,
    // composition : plusieurs composants
    pixels: Vec<Color>,
}

impl Palette {
    pub fn new(size: usize) -> Self {
        Self {
            size,
            next: 0,
            pixels: vec![
                Color {
                    red: 0,
                    green: 0,
                    blue: 0,
                };
                size
            ],
        }
    }

    pub fn add(
        &mut self,
        color: Color,
    ) {
        self.pixels[self.next] = color;
        self.next = (self.next + 1).rem_euclid(self.size);
    }

    pub fn draw(&self) {
        println!("\tsize {}", self.size);
        for c in self.pixels.iter() {
            println!("\t{} {} {}", c.red, c.green, c.blue);
        }
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// composition : plusieurs composants (un autre exemple)
pub struct Image {
    width: usize,
    height: usize,
    pixels: Vec<Color>,
}

impl Image {
    pub fn new_black(
        width: usize,
        height: usize,
    ) -> Self {
        Self {
            width,
            height,
            pixels: vec![
                Color {
                    red: 0,
                    green: 0,
                    blue: 0,
                };
                width * height
            ],
        }
    }

    pub fn draw(&self) {
        println!("\tdimensions {}x{}", self.width, self.height);
        for c in self.pixels.iter() {
            println!("\t{} {} {}", c.red, c.green, c.blue);
        }
    }
}
