// Exemple d'association plus complexe.
// Shop possède plusieurs Item
// Cart est un agrégat de (Shop, ensemble d'Item)
// On fait le choix fort que les shop existeront toujours
// donc on utilise une référence pour exprimer l'association
// mais pour les items c'est trop compliqué et même avec
// toute l'attention possible, le code finale n'assurerait pas
// que le même objet ne soit pas vendu au même moment
// Donc on préfère que les items ont un code pour les distiguer
// et que Cart ne garde que le code des items et pas une référence.
use std::{
    cell::{Ref, RefCell},
    collections::HashMap,
};

pub struct Item {
    pub kind: String,
    pub code: u32,
    pub price: f32,
}
impl Item {
    pub fn new(
        kind: String,
        price: f32,
    ) -> Self {
        Self {
            kind,
            code: 0,
            price,
        }
    }
}
impl std::fmt::Debug for Item {
    fn fmt(
        &self,
        f: &mut std::fmt::Formatter<'_>,
    ) -> std::fmt::Result {
        write!(f, "«{}/{}/{}€»", self.kind, self.code, self.price)
    }
}

pub struct Shop {
    name: String,
    next_code: u32,
    stock: RefCell<HashMap<String, Vec<Item>>>,
}
impl Shop {
    pub fn new(name: String) -> Self {
        Self {
            name,
            next_code: 0,
            stock: RefCell::new(HashMap::new()),
        }
    }
    pub fn name(&self) -> &str {
        self.name.as_str()
    }
    pub fn stock(&self) -> Ref<HashMap<String, Vec<Item>>> {
        self.stock.borrow()
    }
    pub fn receive(
        &mut self,
        mut item: Item,
    ) {
        self.next_code += 1;
        item.code += self.next_code;
        self.stock
            .get_mut()
            .entry(item.kind.clone())
            .or_default()
            .push(item);
    }
    pub fn remove(
        &self,
        code: u32,
    ) -> Result<Item, Box<dyn std::error::Error>> {
        for (_kind, items) in self.stock.borrow_mut().iter_mut() {
            if let Some(pos) = items.iter().position(|i| i.code == code) {
                return Ok(items.remove(pos));
            }
        }
        Err(format!("no code {} in {:?}", code, self.name))?
    }
    pub fn show(&self) -> String {
        let mut txt = format!("{:?}:", self.name);
        for (kind, items) in self.stock.borrow().iter() {
            txt += &format!("\n• {} {:?}", kind, items);
        }
        txt
    }
}

#[derive(Default)]
pub struct Cart<'s> {
    items: Vec<(&'s Shop, u32)>,
}
impl<'s> Cart<'s> {
    pub fn new() -> Self {
        Self { items: Vec::new() }
    }
    pub fn choose_item(
        &mut self,
        shop: &'s Shop,
        kind: &str,
    ) -> Result<(), Box<dyn std::error::Error>> {
        let stock = shop.stock();
        let items = stock.get(kind).ok_or_else(|| {
            format!("unknown kind {:?} in {:?}", kind, shop.name(),)
        })?;
        let item = items.last().ok_or_else(|| {
            format!("no more {:?} in {:?}", kind, shop.name(),)
        })?;
        self.items.push((shop, item.code));
        Ok(())
    }
    pub fn buy(&mut self) -> Vec<Item> {
        let mut items = Vec::new();
        for (shop, code) in self.items.drain(..) {
            if let Ok(item) = shop.remove(code) {
                items.push(item);
            } else {
                println!("no item {} in {:?}", code, shop.name());
            }
        }
        items
    }
}
